######################################################################
# Class to estimate the mixing distribution, then estimate posterior #
######################################################################

import numpy as np
import pandas as pd
from scipy.stats import norm
from patsy import dmatrix
from sklearn.preprocessing import scale
from scipy.optimize import minimize, minimize_scalar
from scipy.stats import norm
from drrank_prior import minimgap, likelihood
import tqdm
from joblib import Parallel, delayed
import multiprocessing
import matplotlib.pyplot as plt
import seaborn as sns

class prior_estimate():
    """
    Class to estimate both prior and posterior distribution.
    Inputs should be a vector of estimates and standard errors previously calculated.
    Arguments:
        thetas: vector of estimates
        s: standard errors of the estimates thetas
        
    """
    def __init__(self, thetas, s, transform=lambda x: x):

        # Transform estimates to lists
        if isinstance(thetas, list):
            self.thetas = np.asarray([thetas]).T
        else:
             self.thetas = thetas
        if isinstance(s, list):
            self.s = np.asarray([s]).T 
        else:
             self.s = s

        # Get the number of units we are ranking
        self.F = len(thetas)

        # Inverse of any transform used to variance stabilize estimates
        self.inv_transform = transform

    def estimate_prior(self, supp_points=5000, spline_order=5, seed=None):
        """
        Estimate the prior distribution G
        Arguments:
            supp_points: support points of theta and z-score distributions (default = 5000)
            spline_order: spline order for our prior distribution estimate, 0.45 for cont, 0.49 for disc efron, 0.0116 for poisson, 0.0031 for balanced poisson (default = 5)
            seed: specify seed number
        """
        print("Estimating prior distribution")
        self.supp_points = supp_points
        self.spline_order = spline_order

        # Calculate support of prior
        supp_theta_min = np.min(self.thetas)
        supp_theta_max = np.max(self.thetas)
        supp_theta = np.linspace(supp_theta_min, supp_theta_max, self.supp_points)
        M = len(supp_theta)
        self.M = M

        # Save the support information
        self.supp_theta_min = supp_theta_min
        self.supp_theta_max = supp_theta_max
        self.supp_theta = supp_theta

        # Calculate P matrix
        s_tilde_big = np.tile(self.s.reshape(-1,1), (1, M))
        thetas_big = np.tile(self.thetas.reshape(-1,1), (1, M))
        supp_theta_big = np.tile(supp_theta.reshape(1, M), (self.F, 1))
        P = (1 / s_tilde_big) * norm.pdf((thetas_big - supp_theta_big) / s_tilde_big)
        self.P = P

        # Report mean and std dev of prior
        mean_ests = np.mean(self.thetas) # Calculate the mean of theta_hat
        sd_ests = np.sqrt(np.var(self.thetas, ddof=1) - np.mean(self.s**2))

        # Save them
        self.mean_ests = mean_ests
        self.sd_ests = sd_ests

        # Estimate the variance-covariance matrix of mean and sd. estimates
        var_mean_ests = np.sum(self.s**2) / self.F**2 
        var_sd_ests = (2 * np.sum(self.s**4) + 4 * np.sum((self.thetas - mean_ests)**2 - self.s**2)) / self.F**2 
        covar_mean_sd_ests = -2 / self.F**2 * np.sum(((self.thetas - mean_ests)**2) * self.s**2) 
        vcv_ests = np.array([[var_mean_ests, covar_mean_sd_ests], [covar_mean_sd_ests, var_sd_ests]]) 
        
        # Save it
        self.vcv_ests = vcv_ests

        # Comput the spline bais
        X = supp_theta.reshape(-1, 1)
        std_spline = 1

        # Use R if possible for consistency with Efron, otherwise use Patsy BSpline
        try:
            # import rpy2's package module
            import rpy2.robjects.packages as rpackages
            import rpy2.robjects as robjects
            import rpy2.robjects.numpy2ri
            rpy2.robjects.numpy2ri.activate()

            # import R's utility package
            splines = rpackages.importr('splines')

            # Get the ns function from R
            ns = robjects.r['ns']

            # Convert our X matrix to an R object
            nr, nc = X.shape
            Xr = robjects.r.matrix(X, nrow=nr, ncol=nc)
            Q = ns(Xr, df = self.spline_order)
        except Exception as e:
            print("Cannot use R to create cubic spline basis, switching to python")
            Q = dmatrix("cr(x, df = T)-1", {"x": X, 'T': self.spline_order}, return_type='dataframe')

        if std_spline > 0:
            # Standardize
            Q = scale(Q)
            Q = np.apply_along_axis(lambda w: w / np.sqrt(np.sum(w * w)), 0, Q)

        # Tune G to match mean and SD
        rng = np.random.default_rng(seed=seed)
        alpha_0 = rng.standard_normal((self.spline_order, 1)) 

        # Setup the solver options
        options_fmin = {
            'disp': False,
            'maxiter':10000,
        }

        print("\nPicking penalzation parameter...")

        # Minimize and get the correct c
        result = minimize_scalar(lambda x: minimgap(x, P, Q, alpha_0, options_fmin, 
                                                    supp_theta, sd_ests, mean_ests, 
                                                    vcv_ests),
                                bounds=(0, 0.1),
                                method='bounded',
                                options={'maxiter':500,
                                         'xatol':1e-4})
        c = result.x

        # Check if the minimization was successful
        if result.success != True:
            raise AssertionError("Optimization was unsuccessful")

        print(f"Using df {self.spline_order} and penalty {c:.8f}")
        
        # minimize and solve our likelihood function
        print("\nOptimizing likelihood...")
        result = minimize(lambda x: likelihood(x, P, Q, c), alpha_0,
                        method='CG', jac=True, 
                        options=options_fmin, tol=1e-12)
        alpha_hat = result.x
        print("Likelihood: {:5.4f}".format(result.fun))
        if result.success != True:
            print("Warning: likelihood may not have converged")
            print("Jacobian squared norm: {}".format(np.sum(np.power(result.jac,2))))

        # Get the estimated g_thetas
        logL, dlogL, g_theta = likelihood(alpha_hat, P, Q, c, optimization = False)

        # Report mean and std dev of estimated theta distribution
        mean_theta = np.sum(supp_theta * g_theta) / np.sum(g_theta)
        sd_theta = np.sqrt((np.sum((supp_theta ** 2) * g_theta) / np.sum(g_theta)) - mean_theta ** 2)

        # Print results
        print(f"\nEstimated mean: {mean_ests:4f}")
        print(f"Estimated standard deviation: {sd_ests:4f}")
        print(f"Prior mean: {mean_theta:4f}")
        print(f"Prior standard deviation: {sd_theta:4f}")

        # Rescale g to sum to one
        g_theta = g_theta / np.sum(g_theta)

        # Save the estimates
        self.prior_g = {'mean_theta': mean_theta, 'sd_theta': sd_theta, 'g_theta': g_theta}

    def plot_estimates(self, g_theta=None, show_plot=True, save_path=None):
        """
        Function to plot an histogram of the estimates
        Arguments:
            g_theta: provide your own estimates of the prior distribution
            show_plot: set to True to display the plot
            save_path: specify the path to save the plot, set to None to avoid saving it
        """

        if g_theta == None:
            try:
                g_theta = self.prior_g['g_theta']
            except:
                raise ValueError("Prior distribution was not estimated, cannot proceed to compute the posteriors")
        else:
            print("Using user-supplied prior distribution")
                

        # Get the estimated prior_g
        g_theta = self.prior_g['g_theta']

        # Scale to match histogram
        g_theta = g_theta/max(g_theta)/3
        support = self.supp_theta.copy()
        thetas = self.thetas.copy()

        # Calculate the limits of our plot
        x_min = min(support)*0.99
        x_max = max(support)*1.01
        y_lim = min([round(max(g_theta),1)+0.2, 1])

        # Plot the distribution
        fig, ax = plt.subplots(figsize=(15, 10))
        sns.histplot(data = thetas, 
                    line_kws = {'alpha': 0.6}, kde = False,
                    binwidth = 0.0025,
                    stat = 'probability',
                    fill = True,
                    alpha = 0.3, common_norm = False, ax = ax)
        sns.lineplot(x = support, y = g_theta, color = 'red')
        sns.despine()
        plt.ylabel('Scaled density / mass', fontsize = 25)
        plt.xlabel(r'$\theta$', fontsize = 25)
        plt.ylim(0,y_lim)
        plt.xlim(x_min, x_max)
        plt.yticks(fontsize = 20)
        plt.xticks(fontsize = 20)
        ax.set_axisbelow(True)
        ax.yaxis.grid(color='gray', alpha = 0.2)
        ax.text(x_max, y_lim*0.84, "Bias-corrected SD: {:4.4f}\nDecon. implied SD: {:4.4f}\nMean: {:4.4f}\nDecon. implied mean: {:4.4f}".format(
            self.sd_ests, self.prior_g['sd_theta'], self.mean_ests, self.prior_g['mean_theta']), fontsize = 20, horizontalalignment = 'right')

        if save_path != None:
            plt.tight_layout()
            plt.savefig(save_path, format='jpg', dpi=300)

        if show_plot:
            plt.show()

    def compute_posterior_distributions(self, g_theta=None):
        # Decide which prior density to use
        if g_theta == None:
            try:
                g_theta = self.prior_g['g_theta']
            except:
                raise ValueError("Prior distribution was not estimated, cannot proceed to compute the posteriors")
        else:
            print("Using user-supplied prior distribution")

        # Compute posterior distribution for each estimate
        post_dist = ((1 / self.s) * norm.pdf((self.thetas - self.supp_theta[:,np.newaxis]) / self.s)) * g_theta[:,np.newaxis]
        post_dist = post_dist / np.sum(post_dist,0)

        self.post_dist = post_dist

    def posterior_features(self, g_theta, alpha=.05):
        # Estimate posterior if necessary
        if self.post_dist is None:
            self.compute_posterior_distributions(g_theta)

        # Compute posterior features
        self.pmean = self.supp_theta.dot(self.post_dist)
        self.pmean_trans = self.inv_transform(self.supp_theta).dot(self.post_dist)

        self.lci = self.supp_theta[np.argmin(np.cumsum(self.post_dist,0) <= alpha/2,0)]
        self.uci = self.supp_theta[np.argmin(np.cumsum(self.post_dist,0) <= 1-alpha/2,0)]

        self.lci_trans = self.inv_transform(self.lci)
        self.uci_trans = self.inv_transform(self.uci)

        self.posterior_df = pd.DataFrame({
            'pmean': self.pmean,
            'pmean_trans': self.pmean_trans,
            'lci': self.lci,
            'uci': self.uci,
            'lci_trans': self.lci_trans,
            'uci_trans': self.uci_trans
        })
    
    def compute_posteriors(self, alpha=.05, g_theta=None):
        # Get posterior distributions
        self.compute_posterior_distributions(g_theta)

        # Get posterior features
        self.posterior_features(g_theta, alpha)

    def compute_pis(self, g_theta=None, ncores=-1, power=0):
        """
        Estimate pairwise loss compoments. When power=0, these are
        simply Prob(unit_i > unit_j | Y, G). When power > 0, these are
        E[max(unit_i - unit_j,0)^power | Y, G], enabling extensions to loss
        functions that weight ranking loss by powers of the cardinal
        difference between i and j.
        Arguments:
            g_theta: if a different estimation of the prior has been used, supply it through g_theta, default is None, i.e. use the estimates from estimates_prior()
            ncores: how many CPUs for parallel processing, default is all CPUs available (ncores = -1)
            power: cardinal weights. Set to zero (default) for pairwise posterior ordering probabilities
                Set to 1 for absolute error weighted loss. Set to 2 for square-weighted loss.
        
        Access the estimated pairwise components calling self.pis()
        """
        if power < 0:
            raise ValueError("Power option must be weakly positive")

        # Estimate posterior if necessary
        if self.post_dist is None:
            self.compute_posterior_distributions(g_theta)

        # Integrand
        gaps = (np.greater.outer(self.supp_theta, self.supp_theta) 
                * (np.subtract.outer(
                        self.inv_transform(self.supp_theta),
                        self.inv_transform(self.supp_theta)) ** power))

        # If ncores is -1, use all CPUs
        if ncores == -1:
            ncores = multiprocessing.cpu_count() 
                
        print(f"\nCalculating pi matrix, using {ncores} cores...")
        
        # Pi calculation wrapper
        def pair_dif(post_dist, gaps, i,j):
            post_prod = post_dist[:,i][:,np.newaxis] * post_dist[:,j][np.newaxis,:]
            return np.sum(gaps * post_prod) / np.sum(post_prod)

        pis = np.zeros((self.F,self.F))
        for i in tqdm.tqdm(range(self.F)):
            def par_wrapper(j):
                # if power == 0, then p_ij = 1 - p_ji
                if power == 0 and i > j and 0 == 1:
                    return 1 - pis[j,i]
                else:
                    return pair_dif(self.post_dist, gaps, i, j)
            pis[i,:] = Parallel(n_jobs=ncores)(delayed(par_wrapper)(j) for j in range(self.F))

        print("...done!")

        self.pis = pis.copy()
        return pis


import numpy as np
from scipy.optimize import minimize
from scipy.stats import norm


def minimgap(c, P, Q, alpha_0, options_fmin, supp_theta, sd_ests, mean_ests, vcv_ests):
    """
    Minmgap estimation of the G prior function
    """
    result = minimize(lambda x: likelihood(x, P, Q, c), alpha_0,
            method='CG', jac=True, options=options_fmin, tol=1e-10)
    alpha_hat = result.x
    logL, dlogL, g_theta = likelihood(alpha_hat, P, Q, c, optimization=False)

    # Report mean and std dev of estimated theta distribution
    mean_theta = np.sum(supp_theta * g_theta) / np.sum(g_theta)
    sd_theta = np.sqrt((np.sum((supp_theta ** 2) * g_theta) / np.sum(g_theta)) - mean_theta ** 2)
    difs = np.array([mean_theta - mean_ests, sd_ests - sd_theta])
    gap = difs.T @ np.linalg.inv(vcv_ests) @ difs

    return gap

def likelihood(alpha, P, Q, c, optimization = True):
    """
    Likelihood function to be maximized for the G estimation
    """

    # Calculate prior
    g = np.exp(Q @ alpha - logsumexp(Q @ alpha))

    # Calculate objective
    logL = -np.sum(np.log(P @ g)) + c * np.sqrt(alpha.T @ alpha)

    # Calculate gradient of objective
    T = len(alpha)
    M = len(P[0])
    F = len(P)
    dg = np.tile(g.T, (T, 1)) * (Q.T - np.repeat((Q.T @ g).reshape(-1, 1), repeats = M, axis = 1))
    dlogL = -(np.sum(np.repeat((1 / (P @ g)).reshape(-1, 1), repeats = T, axis = 1) * (P @ dg.T), axis=0) - (c * alpha / np.sqrt(alpha.T @ alpha)))
    if optimization:
        return logL, dlogL
    else:
        return logL, dlogL, g

def logsumexp(x, axis=None):
    # Compute log(sum(exp(x), axis)) while avoiding numerical underflow.
    # By default, axis = 0 (columns).
    # Written by Michael Chen (sth4nth@gmail.com) in MATLAB.
    if axis is None:
        # Determine which dimension sum will use
        axis = np.argmax(np.array(x.shape) != 1)
        if np.size(axis) == 0:
            axis = 0

    # Subtract the largest in each column/axis
    y = np.max(x, axis=axis)
    x = x - np.expand_dims(y, axis=axis)
    s = y + np.log(np.sum(np.exp(x), axis=axis))
    i = ~np.isfinite(y)
    if np.any(i):
        s[i] = y[i]

    return s


##########################################################
#### Report card estiamtes functions                  ####
##########################################################
import pandas as pd
import numpy as np
import os
import gurobipy as gp
from gurobipy import GRB
from scipy.sparse.csgraph import connected_components
import multiprocessing
from multiprocessing.pool import ThreadPool
import matplotlib.pyplot as plt
from drrank_loss import dp, tau
import tqdm
import time

## function to clean the Pij matrix ##
def clean_data(pi):
    """
    Function to prepare the P_ij data to be fitted with the report card model
    p_ij: a numpy matrix of (n_obs,n_obs) dimension containing the posterior estimates of the probabilities
          P_ij of observation i being more biased than observation j - P_ij = Pr(Theta_i > Theta_j | Y_i, Y_j)
    """
    # Create a copy to not modify data in place
    p_ij = pi.copy()

    # Requires np.array as inputs
    if type(p_ij) is not np.ndarray:
        raise ValueError("Pi matrix must be numpy array.")

    # Fill the diagonal with 0, so no gain / cost to ranking i tied with i
    np.fill_diagonal(p_ij, 0)

    # Check if p_ij = 1-p_ij.T
    if not np.allclose(p_ij + np.eye(p_ij.shape[0]), (1-p_ij).T):
        print("Warrning: Pi matrix does not appear equal to (1 - Pi.T)")
        print("Max deviation {:7.6f}".format(np.max(np.abs(p_ij +  np.eye(p_ij.shape[0]) -  (1-p_ij).T))))

    # Get coordinate representation
    N = p_ij.shape[0]
    i = np.repeat(np.arange(N),N) + 1
    j = np.tile(np.arange(N),N) + 1
    v = np.ravel(p_ij)

    # Make a dataframe out of the coordinate representation of the matrix
    df_p_ij = pd.DataFrame({'i':i,'j':j,'P_ij':v})

    # Get the matrix coordinates
    df_p_ij['idx'] = list(zip(df_p_ij.i, df_p_ij.j))

    # Make a dictionary with keys the matrix coordinates, values the P_ij
    p_ij_dict = {x:v for x, v in zip(df_p_ij['idx'], df_p_ij['P_ij'])}

    # Split a single dictionary into multiple dictionaries
    i_j, Pij = gp.multidict(p_ij_dict)

    return i_j, Pij

## main function ##
def report_cards(i_j, Pij, lamb = None, DR = None, loss = 'binary', save_controls = False, save_dir = "dump", save_name = '_debug', warmstart = True, add_cond = True, FeasibilityTol = 1e-9, IntFeasTol = 1e-9, OptimalityTol = 1e-9):
    """
    Compute the report cards via Gurobi optimization
    Parameters:
    i_j: Coordinates of the Pij
    Pij: Posterior estimates of the probability of observation i being more biased than observation j
    lamb:  Tuning parameter trading off the gains of correctly ranking pairs of observations against the cost of misclassifying them - either lamb or DR must be specified
    DR: Discordance rate (i.e. shares of observation pairs misranked according to their grades) - either lamb or DR must be specified
    save_controls: if True, saves the estimates for debugging purposes (default = False)
    save_dir: if save_controls == True, name for the directory in which the estimates will be saved, if the default directory is used, it creates a folder named "dump" (default = "dump")
    save_name: if save_controls == True, name for the file in which the estimates will be saved (default = "_debug")
    FeasibilityTol: Feasibility Tollerance, Gurobi parameter (default = 1e-9)
    IntFeasTol: Integer feasibility Tollerance, Gurobi parameter (default = 1e-9)
    OptimalityTol: Optimality Tollerance, Gurobi parameter (default = 1e-9)
    """
    # Time execution
    start = time.time()

    # Check that lambda and DR are correctly specified
    if (lamb is not None) & (DR is not None) :
        raise AssertionError("Must supply either lambda or DR, but not both.")
    if (lamb != None):
        if (lamb > 1) or (lamb < 0):
            raise AssertionError("Lambda must be within [0,1].")
    if (DR != None):
        if (DR > 1) or (DR < 0):
            raise AssertionError("DR must be within [0,1].")

    # the model
    with gp.Env() as env, gp.Model(env=env) as model:
        n_firms = max(i_j)[0] + 1

        # Tolearances
        model.Params.FeasibilityTol = FeasibilityTol
        model.Params.IntFeasTol = IntFeasTol
        model.Params.OptimalityTol = OptimalityTol
        
        # control variables
        Dij = model.addVars(i_j, vtype=GRB.BINARY, name="Dij")
        grades = model.addVars(list(range(1,n_firms)),
                lb=1, ub=n_firms-1,
                vtype=GRB.INTEGER, name="grades")

        # Add constraints        
        print("Building constraints...")
        for i in tqdm.tqdm(range(1, n_firms)):
            for j in range(1, n_firms):
                model.addConstr((Dij[(i,j)] == 1) >>
                    (grades[i] - grades[j] >= 1))
                model.addConstr((Dij[(i,j)] == 0) >>
                    (grades[j] - grades[i] >= 0))
        model.update()

        # Objective
        loss_condorcet = -tau(i_j, Pij, Dij)
        if lamb is not None:
            loss = (1-lamb)*dp(i_j, Pij, Dij) - lamb*tau(i_j, Pij, Dij)
        elif DR is not None:
            loss = loss_condorcet
        else:
            raise AssertionError("Must supply either lambda or DR.")

        # warmstart at lambda=1
        if warmstart:
            print("Warm starting at lambda = 1")
            model.setObjective(loss_condorcet, GRB.MINIMIZE)
            model.update()
            model.optimize()
            D_ij_hat_cond = model.getAttr('x', Dij)

        model.setObjective(loss, GRB.MINIMIZE)
        model.update()

        # DR constraint, if necessary
        if DR is not None:
            npairs = np.sum(Pij.values())
            dr_constr = model.addConstr(dp(i_j, Pij, Dij) <= DR*npairs)
        model.update()

        # First optimize() if call fails - need to set NonConvex to 2
        try:
            print("Optimizing model...")
            model.optimize()
        except gp.GurobiError:
            print("...ptimize failed due to non-convexity")
            print("...optimizing with non-convexity")
            # Solve bilinear model
            model.Params.NonConvex = 2
            model.optimize()
        print("...optimized model successfully")

        ## print results
        print("%%%%%%%%%%%%%%%%%%%%%%%%%")
        if lamb is not None:
            print('lambda: %g' % lamb)
        elif DR is not None:
            print('DR: %g' % DR)
        print('Obj: %g' % model.ObjVal)
        print("%%%%%%%%%%%%%%%%%%%%%%%%%")

        ## get results
        D_ij_hat = model.getAttr('x', Dij)

        data_items = D_ij_hat.items()
        data_list = list(data_items)
        df = pd.DataFrame(data_list, columns=['i_j', 'D_ij'])

        # For debugging save for later
        if save_controls==True:
            if save_dir == "dump":
                # Create a folder if the user has not specified any directory
                os.makedirs("dump/", exist_ok=True)
            print("Saving estimates in {}/df_aux_{}_{}.csv for debugging purposes".format(save_dir,lamb, save_name))
            df.to_csv("{}/df_aux_{}_{}.csv".format(save_dir,lamb, save_name), index = False)

        # Get groups
        print("Getting the implied groups...")
        df['obs_idx'] = df.i_j.apply(lambda x: x[0])
        df_groups = df.groupby('obs_idx').D_ij.sum().to_frame().reset_index()
        df_groups['groups'] = df_groups.D_ij.rank(
                method='dense', ascending=False).astype(int)

        print("Solution yields {} total groups".format(df_groups.groups.nunique()))

        if lamb is not None:
            df_groups.rename(columns={'groups': 'grades_lamb{}'.format(lamb)}, inplace=True)
        elif DR is not None:
            df_groups.rename(columns={'groups': 'grades_DR{}'.format(DR)}, inplace=True)

        ## Add condorcet solution
        if add_cond:
            print("Adding condorcet solution")
            if not warmstart:
                model.setObjective(loss_condorcet, GRB.MINIMIZE)
                if DR is not None:
                    model.remove(dr_constr)
                model.update()
                model.optimize()
                D_ij_hat_cond = model.getAttr('x', Dij)

            data_items = D_ij_hat_cond.items()
            data_list = list(data_items)
            df = pd.DataFrame(data_list, columns=['i_j', 'D_ij'])
            df['obs_idx'] = df.i_j.apply(lambda x: x[0])
            cond_groups = df.groupby('obs_idx').D_ij.sum().to_frame().reset_index()
            cond_groups['condorcet_rank'] = cond_groups.D_ij.rank(
                    method='dense', ascending=False).astype(int)

            df_groups = df_groups.merge(cond_groups.drop('D_ij', axis=1), how='left', on='obs_idx')

        print("Time elapased solving LP problem(s): {:4.3f} minutes".format((time.time()-start)/60))
        return df_groups.drop('D_ij',axis=1)
    

## function to fit the ranking model ##
def fit(Pij, lamb, DR = None, save_controls = False, save_dir = "dump", save_name = '_debug', FeasibilityTol = 1e-9, IntFeasTol = 1e-9, OptimalityTol = 1e-9):
    """
    Function to fit the report card model on a Pij matrix of posterior estimates of bias
    Parameters:
    i_j: Coordinates of the Pij
    Pij: Posterior estimates of the probability of observation i being more biased than observation j
    lamb:  Tuning parameter trading off the gains of correctly ranking pairs of observations against the cost of misclassifying them
    DR: Discordance proportion (i.e. shares of observation pairs misranked according to their grades) - either lamb or DR must be specified
    save_controls: if True, saves the estimates for debugging purposes (default = False)
    save_dir: if save_controls == True, name for the directory in which the estimates will be saved, if the default directory is used, it creates a folder named "dump" (default = "dump")
    save_name: if save_controls == True, name for the file in which the estimates will be saved (default = "_debug")
    FeasibilityTol: Feasibility Tollerance, Gurobi parameter (default = 1e-9)
    IntFeasTol: Integer feasibility Tollerance, Gurobi parameter (default = 1e-9)
    OptimalityTol: Optimality Tollerance, Gurobi parameter (default = 1e-9)
    """    
    # Clean the Pij and retrieve the sparse representation
    i_j, Pij = clean_data(Pij)

    # Produce the report cards
    res_df = report_cards(i_j, Pij, lamb, DR, save_controls = save_controls, save_dir = save_dir, save_name = save_name, FeasibilityTol = FeasibilityTol, IntFeasTol = IntFeasTol, OptimalityTol = OptimalityTol)

    return res_df

## function to get different ranking estimations based on a list of lambdas ##
def fit_multiple(Pij, lamb_list, ncores = 1, save_controls = False, save_dir = "dump", save_name = '_debug', FeasibilityTol = 1e-9, IntFeasTol = 1e-9, OptimalityTol = 1e-9):
    """
    Function to fit the report card model on a Pij matrix of posterior estimates of bias
    -> gives back a dataframe with results for various lambda parameters
    Parameters:
    i_j: Coordinates of the Pij
    Pij: Posterior estimates of the probability of observation i being more biased than observation j
    lamb_list:  List of lambdas to be used (lambda: Tuning parameter trading off the gains of correctly ranking pairs of observations against the cost of misclassifying them)
    ncores: Number of cores, if set to -1 uses all the cores available (default = 1)
    save_controls: if True, saves the estimates for debugging purposes (default = False)
    save_dir: if save_controls == True, name for the directory in which the estimates will be saved, if the default directory is used, it creates a folder named "dump" (default = "dump")
    save_name: if save_controls == True, name for the file in which the estimates will be saved (default = "_debug")
    FeasibilityTol: Feasibility Tollerance, Gurobi parameter (default = 1e-9)
    IntFeasTol: Integer feasibility Tollerance, Gurobi parameter (default = 1e-9)
    OptimalityTol: Optimality Tollerance, Gurobi parameter (default = 1e-9)
    """
    # If ncores is -1, use all CPUs
    if ncores == -1:
        ncores = multiprocessing.cpu_count() 

    # Clean the Pij and retrieve the sparse representation
    i_j, Pij = clean_data(Pij)

    # final dataframe to store all results
    n_obs = max(i_j)[0] + 1
    final_df = pd.DataFrame({'obs_idx': range(1, n_obs)})

    def report_cards_helper(x):
        """
        Helper function to run report_cards within a multiprocessing tool
        x: lambda value over which we iterate
        """
        return report_cards(i_j, Pij, x, DR = None, add_cond=False, save_controls = save_controls, save_dir = save_dir, save_name = save_name, FeasibilityTol = FeasibilityTol, IntFeasTol = IntFeasTol, OptimalityTol = OptimalityTol)

    # Compute using multiprocessing to speed up everything
    print("Computing grades with {} cores".format(ncores), flush=True)
    with ThreadPool(ncores) as p:
        dfs = p.map(report_cards_helper, lamb_list)

    # Get the final result dataset
    for d in dfs:
        final_df = final_df.merge(d, on='obs_idx', validate="1:1")

    return final_df


def fig_ranks(ranking, posterior_features, gradecol=None, ylabels=None, show_plot=True, save_path=None, trans=False, ylabel_fontsize=8):
    """
    Function to plot an histogram of the estimates
    Arguments:
        ranking: dataframe with ranking results from drrank.fit()
        posterior_features: posterior features computed from the drrank.prior_estimate() class
        gradecol: column name with the ranking grades, useful when fitting multiple lambdas. If None, looks for the right columns and pick the first one
        ylabels: column in "ranking" providing the names of the observations, if None simply shows the observation number
        show_plot: set to True to display the plot
        save_path: specify the path to save the plot, set to None to avoid saving it
        trans: set to True if we want to plot the transformed posterior means
    """
    # Find smallest lambda that delivers num_groups
    if gradecol is not None:
        group = [x for x in ranking.keys() if gradecol in x][0]
    else:
        group = [x for x in ranking.keys() if 'grades_' in x][0]
    print("Using group {}".format(group))

    # First step - get posterior features:
    # Merge posterior features with the group rankings
    df_plot = pd.concat([ranking, posterior_features], axis = 1)

    # Sort everything
    df_plot = df_plot.sort_values([group, "condorcet_rank"], ascending=False)

    # Get upper and lower bounds
    if trans:
        df_plot["lb"] = df_plot['pmean_trans'] - df_plot['lci_trans']
        df_plot["ub"] = df_plot['uci_trans'] - df_plot['pmean_trans']
        mean_col = 'pmean_trans'
    else:
        df_plot["lb"] = df_plot['pmean'] - df_plot['lci']
        df_plot["ub"] = df_plot['uci'] - df_plot['pmean']
        mean_col = 'pmean'

    # Construct figure
    fig, ax1 = plt.subplots(dpi=400, figsize=(6,7))
    y_range = np.arange(len(df_plot))
    groups = df_plot[group].unique()
    # Prepare labels for each group
    if len(groups) > 6:
        labels = [r'${} \bigstar$'.format(c) for c in range(len(groups),0,-1)]
    else:
        labels = [r'$\bigstar$'*c for c in range(len(groups),0,-1)]

    # Iterate and produce a plot for each group
    mean_lines = []
    for i, gr in enumerate(np.flip(groups)):
        idx = df_plot[group] == gr
        mean_lines += [ax1.errorbar(df_plot[mean_col][idx], y_range[idx],
                        xerr=[df_plot['lb'][idx], df_plot['ub'][idx]],
                        alpha=0.6, ms=1.5, elinewidth=0.6,
                    fmt='o', capsize=1.5, label=labels[i]
                        )]

    plt.grid(axis='y', alpha=0.35, linewidth=0.3)
    plt.legend(mean_lines, labels, loc='lower right', markerscale=1)
    if ylabels != None:
        plt.yticks(y_range, df_plot[ylabels], fontsize=ylabel_fontsize)    
    else:
        plt.yticks(y_range, df_plot['obs_idx'], fontsize=ylabel_fontsize)    

    ax1.set_xlabel('Posterior means')
    ax1.spines['right'].set_visible(False)
    ax1.spines['top'].set_visible(False)

    if save_path != None:
        plt.tight_layout()
        plt.savefig(save_path, format='jpg', dpi=300)
        print("Figure saved!")

    if show_plot:
        plt.show()

    plt.clf()
    plt.close('all')


######################
# Utilities functions#
######################

def isfloat(num):
    """
    Check whether a number is float
    """
    try:
        float(num)
        return True
    except ValueError:
        return False
